/*
 * DTM (teda DEM) -> VRML konvertor
 * (c) Miso Forisek 2003
 *
 * skompiluje pod gcc (linux, djgpp), mozno aj pod inym ;)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BINARY 0
#define ASCII  1
#define binaryNODATA -9999   // niektore binary DEM maju specialnu hodnotu "no data"
                             // ak nechceme pouzit, staci zmenit -9999 na napr. 100000 (>32767)
#define binaryOURNODATA 0    // akou vyskou nahradime "no data"  
#define asciiMAXRIADOK 10000 // najvacsia dlzka riadku v ascii DEM
#define INFTY 10e9           // nekonecno, t.j. viac ako suradnice lub. bodu

////////////////////////////////////////////////////////////////////
//            konverzia (binary elev. map) -> vrml                //
////////////////////////////////////////////////////////////////////

void ConvertBinary(
  char *finName, char *foutName,  // nazov vstupneho a vystupneho suboru
  int iWidth, int iHeight,        // sirka a vyska mapy
  double fDist,                   // vzdialenost 2 susednych policok vstupu 
                                  //   -- len sa tym skaluje vyska policok
  double scale,                   // tymto sa prenasobia rozmery vstupu
  double creaseAngle,             // ficuria z VRML
  int bigEndian)                  // v akom poradi su bajty?

{
  FILE *fin, *fout;
  unsigned char ch[10]; short int vyska; // na nacitanie zo suboru
  double *riadok,*dalsi;          // hodnoty v aktualnom a nasled. riadku vystupu
  int newWidth, newHeight;
  
  // pomocne premenne
  int x,y,z,newx,newy,rc;
  double xvaha,yvaha;
  
  // otvorime subory
  fin=fopen (finName ,"r"); if (fin ==NULL) { printf("File not found.\n"); exit(1); }
  fout=fopen(foutName,"w"); if (fout==NULL) { printf("Cannot create output.\n"); exit(1); }

  // spocitame rozmery vystupu
  if (scale>1.0) scale=1.0; if (scale<=0.0) scale=1.0; 
  newWidth =(int)(iWidth*scale); 
  newHeight=(int)(iHeight*scale); 

  // alokujeme pamat
  riadok=(double *) malloc((newWidth+5)*sizeof(double));
  if (riadok==NULL) { printf("Not enough memory.\n"); exit(1); } 
  dalsi=(double *) malloc((newWidth+5)*sizeof(double));
  if (dalsi==NULL) { printf("Not enough memory.\n"); exit(1); } 
  memset(riadok,0,(newWidth+5)*sizeof(double));
  memset(dalsi,0,(newWidth+5)*sizeof(double));

  // jednoducha scena:
  //   -- jedno svetlo
  //   -- jeden viewpoint
  //   -- jeden shape
  fprintf(fout,"#VRML V2.0 utf8\n\n");
  fprintf(fout,"PointLight {\n  location 100 100 100\n  ambientIntensity 1.0\n  color 1.0, 1.0, 1.0\n}\n\n");
  fprintf(fout,"Viewpoint {\n  position 50 50 400\n  orientation 1.0 0.0 0.0 0.1\n  fieldOfView 0.8\n}\n\n");
  fprintf(fout,"Shape {\n  appearance Appearance {\n    material Material {\n");
  fprintf(fout,"      diffuseColor 0.7 0.7 0.7\n      shininess 0.2\n");
  fprintf(fout,"    }\n  }\n  geometry ElevationGrid {\n");
  
  // rozmery & co. nasho elevation gridu
  fprintf(fout,"    xDimension %d\n    zDimension %d\n",newWidth,newHeight);
  fprintf(fout,"    xSpacing 1\n    zSpacing 1\n    solid FALSE\n");
//  FIXME
//  tu by asi malo byt taketo, ale potom je to nepozeratelne velkeeee
//    fprintf(fout,"    xSpacing %.2f\n    zSpacing %.2f\n    solid FALSE\n",fDist,fDist);
//  a potom este ked vypisujeme hodnoty, nebudeme ich uz delit fDist  
  if (creaseAngle>0.000001) fprintf(fout,"    creaseAngle %.3f\n",creaseAngle);
  fprintf(fout,"    height [");

  // spracujeme vstupny subor
  for (rc=0,y=0;y<iHeight;y++) {
    // precitame riadok zo vstupu
    for (x=0;x<iWidth;x++) {
      fread(ch,2,1,fin); 
      if (bigEndian) { vyska=ch[1]; vyska*=256; vyska+=ch[0]; } else { vyska=ch[0]; vyska*=256; vyska+=ch[1]; }
      if (vyska==binaryNODATA) vyska=binaryOURNODATA;

      // priratame nacitanu vysku k prislusnym polickam aktualneho riadku
      newx=(int)(x*scale); newy=(int)(y*scale); 
      xvaha=1.0; yvaha=1.0;
      if ((((newx+1)/scale)-x)<1.0) xvaha=((newx+1)/scale)-x;
      if ((((newy+1)/scale)-y)<1.0) yvaha=((newy+1)/scale)-y;
      
      riadok[newx  ]+=xvaha*yvaha*vyska;
      riadok[newx+1]+=(1-xvaha)*yvaha*vyska;
      dalsi [newx  ]+=xvaha*(1-yvaha)*vyska;
      dalsi [newx+1]+=(1-xvaha)*(1-yvaha)*vyska;
    }
    
    // ak uz vieme hodnoty dalsieho riadku vystupu, vypiseme ich
    if ((((newy+1)/scale)-y)<=1.00001) {
      fprintf(fout,"\n      # riadok %d",rc++);
      for (z=0;z<newWidth;z++) {
        riadok[z]*=scale*scale; // spocitame priemernu vysku na kazdom policku
        if (!(z%15)) fprintf(fout,"\n      ");      // novy riadok v subore
        fprintf(fout,"%.4f ",riadok[z]*scale/fDist);
      }
      // a skopirujeme nove hodnoty
      memcpy(riadok,dalsi,(newWidth+5)*sizeof(double));
      memset(dalsi,0,(newWidth+5)*sizeof(double));
    }
  }

  // upraceme
  fprintf(fout,"\n    ]\n  }\n}\n");
  free(riadok); free(dalsi);
  fclose(fin); fclose(fout);
}

////////////////////////////////////////////////////////////////////
//            konverzia (ascii elev. map) -> vrml                 //
////////////////////////////////////////////////////////////////////

void ConvertASCII(
  char *finName, char *foutName,  // nazov vstupneho a vystupneho suboru
  int iWidth, int iHeight,        // sirka a vyska mapy
  double scale,                   // tymto sa prenasobia rozmery vstupu
  double creaseAngle)             // ficuria z VRML

{
  FILE *fin, *fout;
  char *vstup;                    // riadok vstupneho suboru
  double *riadok,*dalsi;          // hodnoty v aktualnom a nasled. riadku vystupu
  int newWidth, newHeight;
  int newx,newy;

  // pomocne premenne
  char *pch;
  double X,Y,V;
  double minX,maxX,deltaX,minY,maxY,deltaY;
  int z,row;
  double xvaha,yvaha,fDist;
 
  // alokujeme pamat
  vstup=(char *) malloc((asciiMAXRIADOK+5)*sizeof(char));
  if (vstup==NULL) { printf("Not enough memory.\n"); exit(1); } 
  memset(vstup,0,(asciiMAXRIADOK+5)*sizeof(char));
  
  // 
  // FIRST PASS: najdeme si minimalnu a maximalnu x-ovu suradnicu
  //
  
  fin=fopen (finName ,"r"); if (fin ==NULL) { printf("File not found.\n"); exit(1); }
  minX=minY=INFTY; maxX=maxY=-INFTY;
  while (!feof(fin)){
    fgets(vstup,asciiMAXRIADOK,fin);  fscanf(fin," ");
    pch=strchr(vstup,'#'); if (pch!=NULL) pch[0]=0; // od # dalej je vstup komentar
    // skusime precitat 3 hodnoty
    if (sscanf(vstup,"%lf %lf %lf",&X,&Y,&V)==3){ 
      if (X<minX) minX=X; if (X>maxX) maxX=X; 
      if (Y<minY) minY=Y; if (Y>maxY) maxY=Y; 
    }
  }
  free(vstup);
  fclose(fin);
  deltaX=(maxX-minX)/(iWidth-1); deltaY=(maxY-minY)/(iHeight-1);
  fDist=(deltaX+deltaY)/2;
 
  // 
  // SECOND PASS: citame, spracuvame analogicky ako binarny 
  // predpoklada sa, ze vstup je zoradeny podla y-ovej rastuco
  //

  // otvorime subory
  fin=fopen (finName ,"r"); if (fin ==NULL) { printf("File not found.\n"); exit(1); }
  fout=fopen(foutName,"w"); if (fout==NULL) { printf("Cannot create output.\n"); exit(1); }
  vstup=(char *) malloc((asciiMAXRIADOK+5)*sizeof(char));
  if (vstup==NULL) { printf("Not enough memory.\n"); exit(1); } 

  // spocitame rozmery vystupu
  if (scale>1.0) scale=1.0; if (scale<=0.0) scale=1.0; 
  newWidth =(int)(iWidth*scale); 
  newHeight=(int)(iHeight*scale); 

  // alokujeme pamat
  riadok=(double *) malloc((newWidth+5)*sizeof(double));
  if (riadok==NULL) { printf("Not enough memory.\n"); exit(1); } 
  dalsi=(double *) malloc((newWidth+5)*sizeof(double));
  if (dalsi==NULL) { printf("Not enough memory.\n"); exit(1); } 
  memset(riadok,0,(newWidth+5)*sizeof(double));
  memset(dalsi,0,(newWidth+5)*sizeof(double));

  // jednoducha scena:
  //   -- jedno svetlo
  //   -- jeden viewpoint
  //   -- jeden shape
  fprintf(fout,"#VRML V2.0 utf8\n\n");
  fprintf(fout,"PointLight {\n  location 100 100 100\n  ambientIntensity 1.0\n  color 1.0, 1.0, 1.0\n}\n\n");
  fprintf(fout,"Viewpoint {\n  position 50 50 400\n  orientation 1.0 0.0 0.0 0.1\n  fieldOfView 0.8\n}\n\n");
  fprintf(fout,"Shape {\n  appearance Appearance {\n    material Material {\n");
  fprintf(fout,"      diffuseColor 0.7 0.7 0.7\n      shininess 0.2\n");
  fprintf(fout,"    }\n  }\n  geometry ElevationGrid {\n");
  
  // rozmery & co. nasho elevation gridu
  fprintf(fout,"    xDimension %d\n    zDimension %d\n",newWidth,newHeight);
//  FIXME aj tu by malo byt:
//  fprintf(fout,"    xSpacing %.3f\n    zSpacing %.3f\n    solid FALSE\n",deltaX,deltaY);
  fprintf(fout,"    xSpacing 1\n    zSpacing 1\n    solid FALSE\n");
  if (creaseAngle>0.000001) fprintf(fout,"    creaseAngle %.3f\n",creaseAngle);
  fprintf(fout,"    height [");

  // spracujeme vstupny subor
  row=0; 
  while (!feof(fin)){
    fgets(vstup,asciiMAXRIADOK,fin);  fscanf(fin," ");
    pch=strchr(vstup,'#'); if (pch!=NULL) pch[0]=0; // od # dalej je vstup komentar
    // skusime precitat 3 hodnoty
    if (sscanf(vstup,"%lf %lf %lf",&X,&Y,&V)==3){ 
      // znormujeme X, Y na suradnice podobne tym v binary
      X-=minX; Y-=minY; X/=deltaX; Y/=deltaY;
      newx=(int)(X*scale); newy=(int)(Y*scale); 
      // ak sme presli na novy riadok, stary(e) vypiseme
      while (row<newy) {
        fprintf(fout,"\n      # riadok %d",row);
        for (z=0;z<newWidth;z++) {
          riadok[z]*=scale*scale; // spocitame priemernu vysku na kazdom policku
          if (!(z%15)) fprintf(fout,"\n      ");      // novy riadok v subore
          fprintf(fout,"%.4f ",riadok[z]*scale/fDist);
        }
        // a skopirujeme nove hodnoty
        memcpy(riadok,dalsi,(newWidth+5)*sizeof(double));
        memset(dalsi,0,(newWidth+5)*sizeof(double));
        row++;
      }
      // spracujeme nacitanu trojicu
      xvaha=1.0; yvaha=1.0;
      if ((((newx+1)/scale)-X)<1.0) xvaha=((newx+1)/scale)-X;
      if ((((newy+1)/scale)-Y)<1.0) yvaha=((newy+1)/scale)-Y;
      
      riadok[newx  ]+=xvaha*yvaha*V;
      riadok[newx+1]+=(1-xvaha)*yvaha*V;
      dalsi [newx  ]+=xvaha*(1-yvaha)*V;
      dalsi [newx+1]+=(1-xvaha)*(1-yvaha)*V;
    }
  }
  
  // docitali sme vstup, vypiseme zvysok mapy
  while (row<newHeight) {
    fprintf(fout,"\n      # riadok %d",row);
    for (z=0;z<newWidth;z++) {
      riadok[z]*=scale*scale; // spocitame priemernu vysku na kazdom policku
      if (!(z%15)) fprintf(fout,"\n      ");      // novy riadok v subore
      fprintf(fout,"%.4f ",riadok[z]*scale/fDist);
    }
    // a skopirujeme nove hodnoty
    memcpy(riadok,dalsi,(newWidth+5)*sizeof(double));
    memset(dalsi,0,(newWidth+5)*sizeof(double));
    row++;
  }

  // upraceme
  fprintf(fout,"\n    ]\n  }\n}\n");
  free(vstup); free(riadok); free(dalsi);
  fclose(fin); fclose(fout);
}


////////////////////////////////////////////////////////////////////
//                      help ?                                    //
////////////////////////////////////////////////////////////////////

void ShowHelp(void){
  // sem moze ist popis z README :)
}

////////////////////////////////////////////////////////////////////
//                      MAIN                                      //
////////////////////////////////////////////////////////////////////

int main(int argc, char **argv){
  char finName[1000], foutName[1000]; // premenne na nazov vstup. a vystup. suboru
  int iWidth=4800, iHeight=6000; // mame default hodnoty, ak dostane ine na vstupe, tieto ignoruje
  double fDist=1.0;              // vzdialenost susednych policok vstupu 
  double dScale=1.0;             // aby vysledok nebol privelky, mozeme vstup skalovat
  int fileType=BINARY;           // aky typ suboru ideme konvertovat
  int binaryBigEndian=0;         // je najvyznamnejsi bajt posledny?
  double creaseAngle=0.0,cc;     // ficuria z VRML
  int arg,pom;
  
  finName[0]=0; strcpy(foutName,"output.wrl");
  // parsujeme argumenty programu
  for (arg=1; arg<argc; arg++) {
    if (sscanf(argv[arg],"-x%d",&pom)) { iWidth=pom; continue; }
    if (sscanf(argv[arg],"-width%d",&pom)) { iWidth=pom; continue; }
    
    if (sscanf(argv[arg],"-y%d",&pom)) { iHeight=pom; continue; }
    if (sscanf(argv[arg],"-z%d",&pom)) { iHeight=pom; continue; }
    if (sscanf(argv[arg],"-height%d",&pom)) { iHeight=pom; continue; }
    
    if (sscanf(argv[arg],"-dist%lf",&cc)) { fDist=cc; continue; }
    if (sscanf(argv[arg],"-scale%lf",&cc)) { dScale=cc; continue; }
    if (sscanf(argv[arg],"-crease%lf",&cc)) { creaseAngle=cc; continue; }
    if (!strcmp(argv[arg],"-bigendian")) { binaryBigEndian=1; continue; }
    if (!strcmp(argv[arg],"-binary")) { fileType=BINARY; continue; }
    if (!strcmp(argv[arg],"-ascii")) { fileType=ASCII; continue; }
    if (!strcmp(argv[arg],"-h")) { ShowHelp(); return 0; }
    // inak je to nazov vstupu alebo vystupu
    if (!strlen(finName)) strncpy(finName,argv[arg],997); else strncpy(foutName,argv[arg],997);
  }
  
  switch (fileType) {
    case BINARY: 
      ConvertBinary(finName, foutName, iWidth, iHeight, fDist, dScale, creaseAngle, binaryBigEndian); break;
    case ASCII: 
      ConvertASCII(finName, foutName, iWidth, iHeight, dScale, creaseAngle); break;
  }

  return 0;
}
